import sys

def clear_line():
    """
    Clears the current console line using a carriage return and space padding.
    Adjust the padding length if needed to cover longer lines.
    """
    # Move the cursor back to the beginning of the line
    sys.stdout.write('\r')
    # Overwrite the line with spaces
    sys.stdout.write(' ' * 80)  # Assuming 80 characters is the width of your console
    # Move the cursor back to the beginning again
    sys.stdout.write('\r')

def green_check():
    """
    Prints a green check mark in the console.
    """
    # ANSI escape code for green text
    green_color = '\033[92m'
    # Unicode character for check mark
    check_mark = '\u2714'

    # Print the check mark in green
    return green_color + check_mark

def red_check():
    # ANSI escape code for red text
    red_color = '\033[91m'
    # ANSI escape code to reset color to default
    reset_color = '\033[0m'
    # Unicode character for "X"
    x_mark = '\u2716'

    # Print the "X" in red
    return red_color + x_mark

def reset_color():
    reset_color = '\033[0m'
    return reset_color


class Dataset:
    def __init__(self, dataset_name, data):
        self.name = dataset_name
        self.data = data
    
    def to_json(self):
        return {
            "name": self.name,
            "data": self.data
        }


import requests
import inspect
import time
import json
from .utils.display_util import clear_line, green_check, red_check, reset_color
import time
from urllib.parse import parse_qs, quote
from .utils.source_code import get_source_code_hash, get_json_obj_hash
from .utils.constants import BASE_CLIENT_URL, PROJECT_BASE_URL, MODEL_BASE_URL, RUN_BASE_URL, DATASET_BASE_URL, JUDGE_BASE_URL, EVALUATION_BASE_URL
from .utils.file import upload_data, download_data, read_source_code_as_func


def print_error_message(code):
    try:
        print(error_code_message_map[code])
    except:
        print(error_code_message_map['unexpected_error'])

error_code_message_map = {
    'invalid_api_key': 'Invalid API Key. Please verify your API Key or generate a new one at https://lyra-ml.com Workspace Settings tab.',
    'exceed_storage_limit': '🚧 You\'ve reached your storage limit! To keep everything running smoothly, please visit our Pricing Page to explore upgrade options that suit your needs.',
    'unexpected_error': 'Oops! Something went wrong on our end. Please try again later. If the issue persists, feel free to contact our support team. We\’re here to help!'
}

class Lyra:
    def __init__(self, project_name, API_KEY):
        self.projectId = None
        self.error_code = None
        self.api_key = API_KEY
        self.judges = []
        response = requests.post(PROJECT_BASE_URL, json={'project_name': project_name}, headers={'Authorization': f'Bearer {API_KEY}'})
        if response.status_code == 201:
            body = json.loads(response.text)
            self.projectId = body["project"]["id"]
        else:
            body = json.loads(response.text)
            self.error_code = body["code"]
            print_error_message(self.error_code)


    def _download_dataset(self, dataset_tag):
        # Should get the download url and then download it here.
        response = requests.get(DATASET_BASE_URL + "/" + dataset_tag + "?projectId=" + self.projectId, headers={'Authorization': f'Bearer {self.api_key}'})
        if response.status_code != 200:
            raise ValueError("Failed to download dataset")

        dataset = json.loads(response.text)["dataset"]
        
        parsed_data = parse_qs(download_data(dataset["filename"], self.api_key))
        parsed_data['name'] = parsed_data['name'][0]
        return parsed_data

    def upload_dataset(self, dataset):
        dataset_hash = get_json_obj_hash(dataset.to_json())

        tag_response = requests.post(DATASET_BASE_URL + "/tag", json={
            'name': dataset.name,
            'hash': dataset_hash,
            'projectId': self.projectId,
        }, headers={'Authorization': f'Bearer {self.api_key}'})
        if tag_response.status_code != 200:
            raise ValueError("Failed to get dataset tag")

        tag = json.loads(tag_response.text)["tag"]
        dataset_filename = self.projectId + "_" + tag + ".txt"
        upload_data(dataset_filename, dataset.to_json(), self.api_key)

        dataset_response = requests.post(DATASET_BASE_URL, json={
            'filename': dataset_filename,
            'hash': dataset_hash,
            'name': dataset.name,
            'projectId': self.projectId,
            'tag': tag,
        }, headers={'Authorization': f'Bearer {self.api_key}'})

        if dataset_response.status_code != 200: 
            raise ValueError("Failed to post dataset tag: {tag}")
        return tag


    def add_judge(self, judge):
        # Grab the source code and hash of scoring_rubric and passing_criteria
        scoring_source_code, judge_scoring_hash = get_source_code_hash(judge.scoring_rubric)
        passing_criteria_source_code, judge_passing_hash = get_source_code_hash(judge.passing_criteria)
        
        judge_tag_response = requests.post(JUDGE_BASE_URL + "/tag", json={
            'judge_passing_hash': judge_passing_hash,
            'judge_scoring_hash': judge_scoring_hash,
            'name': judge.name,
            'projectId': self.projectId,
        }, headers={'Authorization': f'Bearer {self.api_key}'})
        
        if judge_tag_response.status_code != 200:
            raise ValueError("Failed to get judge tag")

        judgeTag = json.loads(judge_tag_response.text)["judgeTag"]

        scoring_filename = self.projectId + "_" + judgeTag + "_scoring.txt"
        upload_data(scoring_filename, scoring_source_code, self.api_key)


        passing_filename = self.projectId + "_" + judgeTag + "_passing.txt"
        upload_data(passing_filename, passing_criteria_source_code, self.api_key)

        response = requests.post(JUDGE_BASE_URL, json={
            'judge': json.dumps(judge.to_json()),
            'passing_code_filename': passing_filename,
            'passing_hash': judge_passing_hash,
            'projectId': self.projectId,
            'scoring_code_filename': scoring_filename,
            'scoring_hash': judge_scoring_hash,
            # 'data_size': scoring_source_data_size + passing_data_size,
            'tag': judgeTag,
        }, headers={'Authorization': f'Bearer {self.api_key}'})

        if response.status_code == 201:
            body = json.loads(response.text)
            if body["isNew"]:
                print(f"✔ Judge {judgeTag} created.")
            else:
                print(f"✔ Judge {judgeTag} already exists.")
            
                for item in body["itemsUpdated"]:
                    print(f"✔ Judge {item['field']} field has been updated to: \"{item['value']}\"")
        
    def _get_judge_from_judge_tag(self, judge_tag):
        encoded_judge_tag = quote(judge_tag)
        judge_response = requests.get(JUDGE_BASE_URL + "/" + encoded_judge_tag + "/projects/" + self.projectId, headers={'Authorization': f'Bearer {self.api_key}'})
        if judge_response.status_code != 200:
            raise ValueError("Failed to get judge from judge tag")
        
        judge = json.loads(judge_response.text)["judge"]
        return judge


    def _get_scoring_rubric_func_from_judge(self, judge):
        scoring_source_code = download_data(judge["scoringCodeFilename"], self.api_key)
        return read_source_code_as_func(scoring_source_code, "scoring_rubric")


    def _get_passing_criteria_func_from_judge(self, judge):
        passing_source_code = download_data(judge["passingCodeFilename"], self.api_key)
        return read_source_code_as_func(passing_source_code, "passing_criteria")


    def evaluate_with(self, judge_tag):
        judge = self._get_judge_from_judge_tag(judge_tag)        

        def decorator(func):
            def wrapper(*args, **kwargs):
                dataset = self._download_dataset(judge["dataset_tag"])
                passed_cases = 0
                failed_cases = 0
                i = 0
                scores = []
                for row in dataset["data"]:
                    message = f"Evaluating {dataset['name']} {i}/{len(dataset['data'])}..."
                    print(f"\n{message}", end="", flush=True)
                    outputs = func(row)
                    scoring_rubric = self._get_scoring_rubric_func_from_judge(judge)
                    passing_criteria = self._get_passing_criteria_func_from_judge(judge)

                    score = scoring_rubric(outputs)
                    scores.append({
                        "output": outputs,
                        "score": score,
                    })
                    clear_line()
                    if passing_criteria(score):
                        passed_cases += 1
                        print(f"\r{green_check()} {dataset['name']}[{i}] score: {score} {reset_color()}", end="", flush=True)
                    else:
                        failed_cases += 1
                        print(f"\r{red_check()} {dataset['name']}[{i}] score: {score} {reset_color()}", end="", flush=True)
                    i += 1

                if failed_cases == 0:
                    print(f"\n{green_check()} All cases in dataset passed.")
                else:
                    print(f"\n\n{red_check()} {failed_cases} out of {len(dataset['data'])} failed")

                response = requests.post(EVALUATION_BASE_URL, json={
                    'scores': scores,
                    'judge_tag': judge["tag"],
                    'projectId': self.projectId,
                })

                if response.status_code != 201:
                    raise ValueError("Failed to create evaluation")
                
                return func(*args, **kwargs)
            return wrapper
        return decorator


    def trace(self, func):
        def wrapper(*args, **kwargs):
            if self.error_code:
                return

            params = inspect.signature(func).parameters
            source_code, hash_digest = get_source_code_hash(func)

            start = time.perf_counter()
            result = func(*args, **kwargs)
            latency = time.perf_counter() - start

            # This whole getting model tag and conditionally creation of model could be pulled out separately
            # Perhaps we can make it even more generic by doing the same for evaluations and datasets. 
            model_tag_response = requests.post(MODEL_BASE_URL + "/tag", json={
                'name': func.__name__,
                'hash': hash_digest,
                'projectId': self.projectId
            }, headers={'Authorization': f'Bearer {self.api_key}'})

            if model_tag_response.status_code != 200:
                raise ValueError("Failed to get the next run id")

            model_tag = json.loads(model_tag_response.text)["tag"]
            model_filename = self.projectId + "_" + model_tag + ".txt"
            upload_data(model_filename, source_code, self.api_key)


            param_keys = [key for key in params]
            formatted_args = []
            i = 0
            for arg in args:
                formatted_args.append({
                    'name': param_keys[i],
                    'value': arg,
                    'type': type(arg).__name__,
                })
                i += 1 
            
            formatted_outputs = [{
                'value': result,
                'type': type(result).__name__,
            }]
            
            model_response = requests.post(MODEL_BASE_URL, json={
                'inputs': [{'name': entry['name'], 'type': entry['type']} for entry in formatted_args],
                'modelTag': model_tag,
                'outputs': [{'type': entry['type']} for entry in formatted_outputs],
                'projectId': self.projectId,
            }, headers={'Authorization': f'Bearer {self.api_key}'})
 
            if model_response.status_code != 201:
                raise ValueError("Failed to create or fetch model with model tag: " + model_tag)

            run_creation_response = requests.post(RUN_BASE_URL, json={
                'inputs': formatted_args,
                'latency': latency,
                'modelTag': model_tag,
                'outputs': formatted_outputs,
                'projectId': self.projectId,
                # 'dataSize': data_size,
            }, headers={'Authorization': f'Bearer {self.api_key}'})

            body = json.loads(run_creation_response.text)
            if run_creation_response.status_code == 201:
                print("✨ View run " + body["run"]["id"] + " at " + BASE_CLIENT_URL + "/workspace/" + body["workspaceId"] + "/project/" + self.projectId)
                return result
            else:    
                self.error_code = body["code"]
                print_error_message(self.error_code)

        return wrapper


import requests
import json
from urllib.parse import quote_plus, parse_qs
from .constants import BASE_URL

# TODO: Need to include projectId? Otherwise, users can just download any code source code
def upload_data(filename, data, api_key):
    encoded_filename = quote_plus(filename)
    signed_url_response = requests.get(BASE_URL + "/signed_upload_url?filename=" + encoded_filename, headers={'Authorization': f'Bearer {api_key}'})
    if signed_url_response.status_code != 200:
        raise ValueError("Failed to create signed url.")
    
    signed_url_body = json.loads(signed_url_response.text)
    signed_url = signed_url_body["signedUrl"]
    if type(data) is str:
        upload_response = requests.put(signed_url, data=data.encode('utf-8'), headers={'Content-Type': 'application/octet-stream'})
    if type(data) is dict:
        upload_response = requests.put(signed_url, data=data, headers={'Content-Type': 'application/octet-stream'})

    if upload_response.status_code != 200:
        raise ValueError("Failed to upload data")
    
    return filename

def download_data(filename, api_key):
    encoded_filename = quote_plus(filename)
    download_response = requests.get(BASE_URL + "/signed_download_url?filename=" + encoded_filename, headers={'Authorization': f'Bearer {api_key}'})
    if download_response.status_code != 200:
        raise ValueError("Failed to download " + filename)
    
    download_response_body = json.loads(download_response.text)
    signed_url = download_response_body["signedUrl"]

    headers = {'Content-Type': 'application/octet-stream'}

    content_response = requests.get(signed_url, headers=headers)
    return content_response.text

def read_source_code_as_func(source_code, func_name):
    local_env = {}
    exec(source_code, {}, local_env)

    # Return the function object from the local_env dictionary
    # The name for "scoring_rubric" should be dynamic
    return local_env[func_name]


# import io
# import os
# import firebase_admin
# from firebase_admin import credentials, storage
# import requests
# import json

# dir_path = os.path.dirname(os.path.realpath(__file__))
# # Production configuration: lyraml-firebase-adminsdk-tzq8c-0de38b28e2.json
# # file_path = os.path.join(dir_path, "lyraml-firebase-adminsdk-tzq8c-0de38b28e2.json")

# # Dev configuration: 
# file_path = os.path.join(dir_path, "lyraml-dev-firebase-adminsdk-v4hs3-b945386918.json")
# cred = credentials.Certificate(file_path)

# # Production configuration: 
# # firebase_admin.initialize_app(cred, {
# #     'storageBucket': 'lyraml.appspot.com'
# # })

# # Dev configuration
# firebase_admin.initialize_app(cred, {
#     'storageBucket': 'lyraml-dev.appspot.com',
# })

# # def upload_file(local_file_path, file_name_in_firebase):
# #     """Uploads a file to Firebase Storage"""
# #     bucket = storage.bucket() # Fetches the default bucket
# #     blob = bucket.blob(file_name_in_firebase)
    
# #     # New file uploaded
# #     blob.upload_from_filename(local_file_path)
    
# #     blob.make_public() # TODO: Need to change this later on
# #     # Returns the public url
# #     return blob.public_url

# # # Example usage
# # local_file_path = './corgi-test.png'
# # file_name_in_firebase = 'corgi-test.png' # You can specify folder structure 
# # public_url = upload_file(local_file_path, file_name_in_firebase)
# # print(f"Uploaded file accessible at {public_url}")

# def upload_data_to_firebase(data, file_name_in_firebase):
#     print("type of data: ", type(data))
#     bucket = storage.bucket()
#     data_size = len(data)
#     blob = bucket.blob(file_name_in_firebase)
#     blob.upload_from_string(data, content_type='application/octet-stream')
    
#     # TODO: Need to change this to some type of signing logic
#     blob.make_public()

#     return blob.public_url, data_size


# def upload_text_to_firebase(text_data, file_name_in_firebase):
#     byte_stream = io.BytesIO(text_data.encode('utf-8'))

#     bucket = storage.bucket()
#     encoded_data = text_data.encode('utf-8')
#     byte_stream = io.BytesIO(encoded_data)
#     data_size = len(encoded_data)
#     blob = bucket.blob(file_name_in_firebase)
#     blob.upload_from_file(byte_stream, content_type='text/plain')
    
#     # TODO: Need to change this to some type of signing logic
#     blob.make_public()

#     return blob.public_url, data_size


# def download_json_file(url):
#     """Download a file from a Firebase URL and save it locally."""
#     response = requests.get(url)
#     response.raise_for_status()  # This will raise an exception for HTTP errors

#     return json.loads(response.content)


# def download_file(url):
#     """Download a file from a Firebase URL and save it locally."""
#     response = requests.get(url)
#     response.raise_for_status()  # This will raise an exception for HTTP errors

#     return response.content



import inspect
import hashlib
import json

def get_source_code_hash(func):
    source_code = inspect.getsource(func)
    hasher = hashlib.sha256()
    hasher.update(source_code.encode('utf-8'))
    hash_digest = hasher.hexdigest()
    return source_code, hash_digest


def get_json_obj_hash(obj):
    serialized_obj = json.dumps(obj).encode('utf-8')
    hasher = hashlib.sha256()
    hasher.update(serialized_obj)
    hash_digest = hasher.hexdigest()
    return hash_digest


# BASE_URL="http://localhost:8080"
BASE_CLIENT_URL="http://localhost:3000"
BASE_URL="https://lyraml-api-2-7c3apmua4a-uc.a.run.app"
PROJECT_BASE_URL = BASE_URL + "/projects"
RUN_BASE_URL = BASE_URL + "/runs"
MODEL_BASE_URL = BASE_URL + "/models"
DATASET_BASE_URL = BASE_URL + "/datasets"
JUDGE_BASE_URL = BASE_URL + "/judges"
EVALUATION_BASE_URL = BASE_URL + "/evaluations"